"""Contains generic controllers that can be used for Kubernetes controllers."""
from typing import Optional

from lab_orchestrator_lib.kubernetes.api import APIRegistry, NamespacedApi, NotNamespacedApi
from lab_orchestrator_lib.template_engine import TemplateEngine


class KubernetesController:
    """Base class for Kubernetes controllers.

    All Kubernetes controllers should inherit from KubernetesController. This class adds a method to get templates that
    are used for the resources.

    :param template_file: This file is a class attribute and should be overwritten by the class. It should have the
                          filename of a template in `lab_orchestrator_lib/templates`.
    """

    template_file = None

    def __init__(self, registry: APIRegistry, template_engine: Optional[TemplateEngine] = None):
        """Initializes a KubernetesController.

        :param registry: The APIRegistry that should be used.
        :param template_engine: Optional template engine that is used to read the templates. If set to None the default
                                is `lab_orchestrator_lib.template_engine.TemplateEngine`.
        """
        self.registry = registry
        if template_engine is None:
            self.template_engine = TemplateEngine()
        else:
            self.template_engine = template_engine

    def _get_template(self, template_data) -> str:
        """Returns a template filled with the template data.

        :param template_data: Data that should be inserted into the template.
        :return: YAML str template with the data filled.
        """
        return self.template_engine.replace_template(template=self.template_file, data=template_data)


class NamespacedController(KubernetesController):
    """Abstract base controller for namespaced resources.

    All Kubernetes controllers that uses namespaced api resources should inherit from this class.
    """

    def _api(self) -> NamespacedApi:
        """Gives an instance of the namespaced api that is used in this controller.

        :return: An instance of the namespaced api.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_list(self, namespace) -> str:
        """Gives a list of all objects in the namespace.

        :param namespace: Namespace where to get the objects from.
        :return: A YAML string that contains all objects.
        """
        return self._api().get_list(namespace)

    def get(self, namespace, identifier) -> str:
        """Gives a specific object in the namespace.

        :param namespace: Namespace of the object.
        :param identifier: Identifier of the object.
        :return: A YAML string that contains the object.
        """
        return self._api().get(namespace, identifier)

    def delete(self, namespace, identifier) -> str:
        """Deletes a specific object in the namespace.

        :param namespace: Namespace of the object.
        :param identifier: Identifier of the object.
        :return: A string that contains the deletion status.
        """
        return self._api().delete(namespace, identifier)


class NotNamespacedController(KubernetesController):
    """Abstract base controller for not namespaced resources.

    All Kubernetes controllers that uses not namespaced api resources should inherit from this class.
    """

    def _api(self) -> NotNamespacedApi:
        """Gives an instance of the not namespaced api that is used in this controller.

        :return: An instance of the not namespaced api.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_list(self):
        """Gives a list of all objects.

        :return: A YAML string that contains all objects.
        """
        return self._api().get_list()

    def get(self, identifier):
        """Gives a specific object.

        :param identifier: Identifier of the object.
        :return: A YAML string that contains the object.
        """
        return self._api().get(identifier)

    def delete(self, identifier):
        """Deletes a specific object.

        :param identifier: Identifier of the object.
        :return: A string that contains the deletion status.
        """
        return self._api().delete(identifier)



"""Contains a generic controller that can be used for adapters."""
from typing import Generic, List, TypeVar, Any

from lab_orchestrator_lib.model.model import DockerImage, Lab, LabInstance

Adapter = TypeVar('Adapter')
LibModelType = TypeVar('LibModelType', DockerImage, Lab, LabInstance)  # subclasses of Model


class AdapterController(Generic[Adapter, LibModelType]):
    """Generic controller for adapter classes.

    This generic class implements some of the methods that are needed for a adapter controller. Whats missing is the
    create method and some specific methods that can't be abstracted.
    """

    def __init__(self, adapter: Adapter):
        """Initializes an adapter controller.

        :param adapter: The adapter of the controller.
        """
        self.adapter = adapter

    def get_all(self) -> List[LibModelType]:
        """Gives all objects of the adapter.

        :return: A list of all objects of the adapter.
        """
        return self.adapter.get_all()

    def get(self, identifier) -> LibModelType:
        """Gives a specific object of the adapter.

        :param identifier: The identifier of the object.
        :return: The specific object.
        """
        return self.adapter.get(identifier)

    def get_by_attr(self, attr, value) -> LibModelType:
        """Gives a specific object of the adapter.

        :param attr: The attribute name that should be used to filter.
        :param value: The value of the attribute that should be filtered.
        :return: The specific object.
        """
        return self.adapter.get_by_attr(attr, value)

    def delete(self, identifier) -> None:
        """Deletes a specific object of the adapter.

        :param identifier: The identifier of the object.
        :return: None
        """
        return self.adapter.delete(identifier)

    def save(self, obj: LibModelType) -> LibModelType:
        """Saves changes of the object to the database.

        :param obj: The object object that contains changes.
        :return: The object.
        """
        return self.adapter.save(obj)

    def filter(self, **kwargs) -> LibModelType:
        """Filters the objects of the adapter and returns the first object that matches the filter criteria.

        The database should be filtered by the attributes and belonging values that are given in the kwargs dictionary.

        :param kwargs: A dictionary with filters.
        :return: The first object that matches the filters.
        """
        return self.adapter.filter(**kwargs)


"""Contains the dataclasses that are used in this project."""
from typing import Union

from lab_orchestrator_lib.custom_exceptions import ValidationError


Identifier = Union[str, int]


class Model:
    """Abstract base class that is used for all classes that should be saved in a database."""

    def __init__(self, primary_key: Identifier):
        """Initializes a model object.

        :param primary_key: A unique value to identify the object.
        """
        self.primary_key = primary_key


class User(Model):
    """A User of the library."""

    def __init__(self, primary_key: Identifier):
        """Initializes a user object.

        :param primary_key: A unique value to identify the user.
        """
        super().__init__(primary_key)


class DockerImage(Model):
    """Link to a Docker Image that contains a VM image.

    A docker image object is a link to a docker image that contains a VM image. This is used to create labs. If your
    image is in docker hub the link only needs to contain `username/reponame:version` and no `https://...` stuff.
    """
    def __init__(self, primary_key: Identifier, name: str, description: str, url: str):
        """Initializes a docker image object.

        :param primary_key: A unique value to identify the object.
        :param name: The name of the docker image. (max. 32 chars)
        :param description: A short description of the docker image. (max. 128 chars)
        :param url: The url to the image. (max. 256 chars)
        :raise ValidationError: if one of the parameters has an invalid value.
        """
        super().__init__(primary_key)
        if len(name) > 32:
            raise ValidationError("name is longer than 32 characters.")
        if len(description) > 128:
            raise ValidationError("description is longer than 128 characters.")
        if len(url) > 256:
            raise ValidationError("url is longer than 256 characters.")
        self.name = name
        self.description = description
        self.url = url


class Lab(Model):
    """Lab is a combination of VMs that can be started.

    When you start a lab a lab_instance will be created and the VMs are started in Kubernetes. A lab is used
    to combine VMs in a scenario.
    """

    def __init__(self, primary_key: Identifier, name: str, namespace_prefix: str, description: str,
                 docker_image_id: Identifier, docker_image_name: str):
        """Initializes a lab object.

        :param primary_key: A unique value to identify the object.
        :param name: The name of the lab. (max. 32 chars)
        :param namespace_prefix: A prefix that is used in the namespace in Kubernetes where the VMs are started. (max. 32 chars)
        :param description: A short description of the docker image. (max. 128 chars)
        :param docker_image_id: The id of the docker image that should be used in this lab.
        :param docker_image_name: The name of the VM. (used when connecting to the VM) (max. 32 chars)
        :raise ValidationError: if one of the parameters has an invalid value.
        """
        super().__init__(primary_key)
        if len(name) > 32:
            raise ValidationError("name is longer than 32 characters.")
        if len(namespace_prefix) > 32:
            raise ValidationError("namespace_prefix is longer than 32 characters.")
        if len(description) > 128:
            raise ValidationError("description is longer than 128 characters.")
        if len(docker_image_name) > 32:
            raise ValidationError("docker_image_name is longer than 32 characters.")
        self.name = name
        self.namespace_prefix = namespace_prefix
        self.description = description
        self.docker_image_id = docker_image_id
        self.docker_image_name = docker_image_name


class LabInstance(Model):
    """A lab instance is a lab that is started by a user.

    A lab instance is linked to many Kubernetes resources. Lab instances are created by the controllers when you start
    a lab. When you create them by your own you're probably doing something wrong.
    """

    def __init__(self, primary_key: Identifier, lab_id: Identifier, user_id: Identifier):
        """Initializes a lab instance object.

        :param primary_key: A unique value to identify the object.
        :param lab_id: The id of the lab that is started.
        :param user_id: The id of the user that has started the lab.
        """
        super().__init__(primary_key)
        self.lab_id = lab_id
        self.user_id = user_id


class LabInstanceKubernetes(Model):
    """A lab instance with a token.

    Doesn't need any adapter and should not be saved in the database. This is used to return the JWT access token when
    the lab is started.
    """
    def __init__(self, primary_key: Identifier, lab_id: Identifier, user_id: Identifier, jwt_token: str):
        """Initializes a lab instance kubernetes object.

        :param primary_key: A unique value to identify the object.
        :param lab_id: The id of the lab that is started.
        :param user_id: The id of the user that has started the lab.
        :param jwt_token: JWT token that can be used to access the VMs in this lab instance.
        """
        super().__init__(primary_key)
        self.lab_id = lab_id
        self.user_id = user_id
        self.jwt_token = jwt_token





"""Contains all implementations of controllers.

In the documentation of this module "you" refers to the developer that uses this library, "your program" means the part
of software that uses the lab orchestrator lib and "our library" refers to the lab orchestrator lib.

When you use this library there are some controllers that you need to use to create resources and some that you don't
need to use. The documentation of the controllers gives you specific information about this.
"""

from typing import List, Optional

from lab_orchestrator_lib.template_engine import TemplateEngine
from lab_orchestrator_lib_auth.auth import generate_auth_token, LabInstanceTokenParams
from lab_orchestrator_lib.controller.adapter_controller import AdapterController
from lab_orchestrator_lib.controller.kubernetes_controller import NamespacedController, NotNamespacedController
from lab_orchestrator_lib.database.adapter import DockerImageAdapterInterface, LabAdapterInterface, \
    LabInstanceAdapterInterface, UserAdapterInterface
from lab_orchestrator_lib.kubernetes.api import NotNamespacedApi, NamespacedApi, APIRegistry
from lab_orchestrator_lib.model.model import DockerImage, Lab, LabInstance, Identifier, User, LabInstanceKubernetes


class UserController:
    """User controller.

    This is the only controller that has no create, delete or save methods because the lab orchestrator lib doesn't create users. It's
    part of the program that uses this library to manage the users. That means you can create, update and delete users
    by your own without using this controller.
    """

    def __init__(self, adapter: UserAdapterInterface):
        """Initializes a user controller.

        :param adapter: User adapter that should be used.
        """
        self.adapter = adapter

    def get_all(self) -> List[User]:
        """Gives all objects of the adapter.

        :return: A list of all objects of the adapter.
        """
        return self.adapter.get_all()

    def get(self, identifier: Identifier) -> User:
        """Gives a specific object of the adapter.

        :param identifier: The identifier of the object.
        :return: The specific object.
        """
        return self.adapter.get(identifier)


class NamespaceController(NotNamespacedController):
    """Controller of Kubernetes Namespaces.

    When you need to create, delete or get namespaces in your program you need to use this controller. Usually you don't
    need to create, get or delete namespaces by your own.
    """

    template_file = "namespace_template.yaml"

    def _api(self) -> NotNamespacedApi:
        """Gives an instance of the namespace api.

        :return: An instance of the namespace api.
        """
        return self.registry.namespace

    def create(self, namespace):
        """Creates a new namespace.

        :param namespace: The name of the namespace.
        :return: YAML str of the namespace.
        """
        template_data = {'namespace': namespace}
        data = self._get_template(template_data)
        return self._api().create(data)


class NetworkPolicyController(NamespacedController):
    """Controller of Kubernetes Namespaces.

    When you need to create, delete or get network policies in your program you need to use this controller. Usually you
    don't need to create, get or delete network policies by you own.
    """

    template_file = 'network_policy_template.yaml'

    def _api(self) -> NamespacedApi:
        """Gives an instance of the network policy api.

        :return: An instance of the network policy api.
        """
        return self.registry.network_policy

    def __init__(self, registry: APIRegistry, template_engine: Optional[TemplateEngine] = None):
        """Initializes a network policy controller.

        :param registry: The APIRegistry that should be used.
        :param template_engine: The template engine that should be used. If none: a default one is used.
        """
        super().__init__(registry, template_engine)
        self.default_name = "allow-same-namespace"

    def create(self, namespace):
        """Creates a new network policy.

        :param namespace: The name of the namespace where the network policy should be created.
        :return: YAML str of the network policy.
        """
        template_data = {'namespace': namespace, 'network_policy_name': self.default_name}
        data = self._get_template(template_data)
        return self._api().create(namespace, data)


class DockerImageController(AdapterController):
    """Docker image controller.

    This controller is used by the library to get access to docker images. When you want to add new docker images, get
    or delete old ones you can do it directly without using this controller.
    """

    def __init__(self, adapter: DockerImageAdapterInterface):
        """Initializes a docker image controller.

        :param adapter: The docker image adapter that is used to connect to the database.
        """
        super().__init__(adapter)

    def create(self, name, description, url) -> DockerImage:
        """Creates a new docker image.

        :param name: Name of the docker image.
        :param description: Description of the docker image.
        :param url: Url of the docker image.
        :return: The created docker image.
        """
        return self.adapter.create(name, description, url)


class LabController(AdapterController):
    """Lab controller.

    This controller is used by the library to get access to labs. When you want to add new labs, get
    or delete old ones you can do it directly without using this controller.
    """

    def __init__(self, adapter: LabAdapterInterface):
        """Initializes a lab controller.

        :param adapter: The lab adapter that is used to connect to the database.
        """
        super().__init__(adapter)

    def create(self, name: str, namespace_prefix: str, description: str, docker_image_id: Identifier,
               docker_image_name: str) -> Lab:
        """Creates a new lab.

        :param name: The name of the lab.
        :param namespace_prefix: The namespace prefix of the lab.
        :param description: The description of the lab
        :param docker_image_id: The docker image id of the lab.
        :param docker_image_name: The docker image name of the lab.
        :return: The created docker image.
        """
        return self.adapter.create(name=name, namespace_prefix=namespace_prefix, description=description,
                                   docker_image_id=docker_image_id, docker_image_name=docker_image_name)


class VirtualMachineInstanceController(NamespacedController):
    """Controller of KubeVirts VMIs.

    When you need to create, delete or get VMIs in your program you need to use this controller. Usually you don't need
    to create or delete VMIs by your own. Starting a VMIs is done automatically when you create a lab instance.
    """

    template_file = 'vmi_template.yaml'

    def __init__(self, registry: APIRegistry, namespace_ctrl: NamespaceController,
                 docker_image_ctrl: DockerImageController,
                 template_engine: Optional[TemplateEngine] = None):
        """Initializes a virtual machine instance controller.

        :param registry: APIRegistry that should be used.
        :param namespace_ctrl: Namespace controller that should be used.
        :param docker_image_ctrl: Docker image controller that should be used.
        :param template_engine: The template engine that should be used. If none: a default one is used.
        """
        super().__init__(registry, template_engine)
        self.namespace_ctrl = namespace_ctrl
        self.docker_image_ctrl = docker_image_ctrl

    def _api(self) -> NamespacedApi:
        """Gives an instance of the vmi api.

        :return: An instance of the vmi api.
        """
        return self.registry.virtual_machine_instance

    def create(self, namespace, lab: Lab):
        """Creates a new virtual machine instance.

        :param namespace: Namespace of the virtual machine instance.
        :param lab: Lab that should be started.
        :return: YAML str of the created virtual machine instance.
        """
        docker_image = self.docker_image_ctrl.get(lab.docker_image_id)
        template_data = {"cores": 3, "memory": "3G",
                         "vm_image": docker_image.url, "vmi_name": lab.docker_image_name,
                         "namespace": namespace}
        data = self._get_template(template_data)
        return self._api().create(namespace, data)

    def get_list_of_lab_instance(self, lab_instance: LabInstance, lab_ctrl: LabController):
        """Gives a list of virtual machine instances that belongs to a specific lab instance.

        :param lab_instance: The lab instance.
        :param lab_ctrl: The lab controller that is used to get the namespace.
        :return: A list of VMIs that belong to this lab instance.
        """
        namespace_name = LabInstanceController.get_namespace_name(lab_instance, lab_ctrl)
        namespace = self.namespace_ctrl.get(namespace_name)
        return self.get_list(namespace_name)

    def get_of_lab_instance(self, lab_instance: LabInstance, virtual_machine_instance_id,
                            lab_ctrl: LabController):
        """Gives a specific of virtual machine instance that belongs to a specific lab instance.

        :param lab_instance: The lab instance.
        :param virtual_machine_instance_id: The id of the vmi.
        :param lab_ctrl: The lab controller that is used to get the namespace.
        :return: The specific VMI.
        """
        namespace_name = LabInstanceController.get_namespace_name(lab_instance, lab_ctrl)
        namespace = self.namespace_ctrl.get(namespace_name)
        return self.get(namespace_name, virtual_machine_instance_id)


class LabInstanceController(AdapterController):
    """Controller of lab instances.

    When you need to create, get or delete VMIs in your program you need to use this controller. Creating a new lab
    instance will automatically create a namespace, a network policy and start all VMIs of the referred lab. You should
    not delete a lab instance by your own, because then the namespace, network policy and the VMIs won't stop running
    in Kubernetes.
    """

    def __init__(self,
                 adapter: LabInstanceAdapterInterface,
                 virtual_machine_instance_ctrl: VirtualMachineInstanceController,
                 namespace_ctrl: NamespaceController,
                 lab_ctrl: LabController,
                 network_policy_ctrl: NetworkPolicyController,
                 user_ctrl: UserController,
                 secret_key: str):
        """Initializes a lab instance controller.

        :param adapter: The lab instance adapter that is used to connect to the database.
        :param virtual_machine_instance_ctrl: The virtual machine instance controller that should be used.
        :param namespace_ctrl: The namespace controller that should be used.
        :param lab_ctrl: The lab controller that should be used.
        :param network_policy_ctrl: The network policy controller that should be used.
        :param user_ctrl: The user controller that should be used.
        :param secret_key: The secret key that should be used to create JWT tokens.
        """
        super().__init__(adapter)
        self.virtual_machine_instance_ctrl = virtual_machine_instance_ctrl
        self.namespace_ctrl = namespace_ctrl
        self.lab_ctrl = lab_ctrl
        self.network_policy_ctrl = network_policy_ctrl
        self.user_ctrl = user_ctrl
        self.secret_key = secret_key

    @staticmethod
    def get_namespace_name(lab_instance: LabInstance, lab_ctrl: LabController) -> str:
        """Returns the namespace name where the resources of a lab instances are created.

        The namespace name is generated by a combination of the labs namespace prefix, the user id and the lab instance
        id. This namespace name is unique for every lab instance.

        :param lab_instance: The lab instance from which you want the namespace name.
        :param lab_ctrl: The lab controller that should be used.
        :return: The name of the namespace.
        """
        lab = lab_ctrl.get(lab_instance.lab_id)
        return LabInstanceController.gen_namespace_name(lab, lab_instance.user_id, lab_instance.primary_key)

    @staticmethod
    def gen_namespace_name(lab: Lab, user_id, lab_instance_id) -> str:
        """Generates the namespace name where the resources of a lab instances are created.

        The namespace name is generated by a combination of the labs namespace prefix, the user id and the lab instance
        id. This namespace name is unique for every lab instance.

        :param lab: The lab that is started.
        :param user_id: The user that starts this lab.
        :param lab_instance_id: The id of the lab instance.
        :return: The name of the namespace.
        """

        return f"{lab.namespace_prefix}-{user_id}-{lab_instance_id}"

    def create(self, lab_id: Identifier, user_id: Identifier) -> LabInstanceKubernetes:
        """Creates a lab instance.

        Creating a lab instance is equivalent to starting a lab for a user. This contains creating a namespace, a
        network policy in the namespace and one or more virtual machine instances. In addition to this a JWT token
        will be created with that the user is able to connect to the virtual machine instances through the LabVNC.

        :param lab_id: The id of the lab.
        :param user_id: The id of the user.
        :return: Returns a lab instance kubernetes object.
        :raise Exception: if parameters are invalid.
        """
        lab = self.lab_ctrl.get(lab_id)
        if lab is None:
            # TODO sinnvolle exception werfen
            raise Exception
        user = self.user_ctrl.get(user_id)
        if user is None:
            # TODO sinnvolle exception werfen
            raise Exception
        lab_instance = self.adapter.create(lab_id=lab_id, user_id=user_id)
        # create namespace
        namespace_name = LabInstanceController.gen_namespace_name(lab, user_id, lab_instance.primary_key)
        namespace = self.namespace_ctrl.create(namespace_name)
        # TODO fix response code
        # TODO log if deletion doesn't work
        #if namespace.response_code != 0:
        #    self.adapter.delete(lab_instance.primary_key)
        #    raise Exception
        # create network policy
        network_policy = self.network_policy_ctrl.create(namespace_name)
        #if network_policy.response_code != 0:
        #    self.adapter.delete(lab_instance.primary_key)
        #    self.namespace_ctrl.delete(namespace_name)
        #    raise Exception
        # create vmi
        vmi = self.virtual_machine_instance_ctrl.create(namespace_name, lab)
        #if vmi.response_code != 0:
        #    self.adapter.delete(lab_instance.primary_key)
        #    self.namespace_ctrl.delete(namespace_name)
        #    raise Exception
        lab_instance_token_params = LabInstanceTokenParams(lab_id, lab_instance.primary_key, namespace_name,
                                                           [lab.docker_image_name])
        token = generate_auth_token(user_id=user_id, lab_instance_token_params=lab_instance_token_params,
                                    secret_key=self.secret_key)
        return LabInstanceKubernetes(primary_key=lab_instance.primary_key, lab_id=lab_id, user_id=user_id, jwt_token=token)

    def delete(self, lab_instance: LabInstance) -> None:
        """Deletes a lab instance.

        This also deletes the created namespace with all resources that are contained in this namespace.

        :param lab_instance: The lab instance that should be deleted.
        :return: None
        """
        lab = self.lab_ctrl.get(lab_instance.lab_id)
        namespace_name = LabInstanceController.gen_namespace_name(lab, lab_instance.user_id, lab_instance.primary_key)
        # this also deletes VMIs and all other resources in the namespace
        self.namespace_ctrl.delete(namespace_name)
        # now delete local object
        super().delete(lab_instance.primary_key)

    def get_list_of_user(self, user: User):
        """Gives a list of lab instances that belong to a specific user.

        :param user: The user that belongs to the lab instances.
        :return: A list of lab instances that belongs to the user.
        """
        # TODO list instead of item
        lab_instances = self.adapter.filter(user_id=user.primary_key)
        return lab_instances

    def save(self, obj: LabInstance) -> LabInstance:
        """Removes the inherited save method, because lab instances can't be changed.

        :raise Exception: Always.
        """
        raise Exception("LabInstances can't be mutated.")



"""Maps the Kubernetes API."""

import logging
from abc import ABC
from typing import Dict, Type, Callable, Union, Optional, Any

import requests

_API_EXTENSIONS_NAMESPACED: Dict[str, Type['NamespacedApi']] = {}
_API_EXTENSIONS_NOT_NAMESPACED: Dict[str, Type['NotNamespacedApi']] = {}


NamespacedApiDecorator = Callable[[Type['NamespacedApi'], ], Type['NamespacedApi']]
NotNamespacedApiDecorator = Callable[[Type['NotNamespacedApi'], ], Type['NotNamespacedApi']]


def add_api_namespaced(name: str) -> NamespacedApiDecorator:
    """Decorator that adds a namespaced api class to the APIRegistry.

    When adding your class to the APIRegistry you later are able to get an instance of the class through an attribute of
    an APIRegistry object.

    :param name: Name of the attribute in the APIRegistry where you should get the API.
    :return: Decorator that adds a namespaced api class to the APIRegistry.
    """

    def inner(cls: Type[NamespacedApi]) -> Type[NamespacedApi]:
        _API_EXTENSIONS_NAMESPACED[name] = cls
        return cls
    return inner


def add_api_not_namespaced(name: str) -> NotNamespacedApiDecorator:
    """Decorator that adds a not namespaced api class to the APIRegistry.

    When adding your class to the APIRegistry you later are able to get an instance of the class through an attribute of
    an APIRegistry object.

    :param name: Name of the attribute in the APIRegistry where you should get the API.
    :return: Decorator that adds a not namespaced api class to the APIRegistry.
    """
    def inner(cls: Type[NotNamespacedApi]) -> Type[NotNamespacedApi]:
        _API_EXTENSIONS_NOT_NAMESPACED[name] = cls
        return cls
    return inner


class Proxy:
    """This proxy is used to make requests to the Kubernetes API.

    This proxy adds authentication headers and checks the SSL certificates.
    """

    def __init__(self, base_uri: str, service_account_token: Optional[str] = None,
                 cacert: Optional[str] = None, insecure_ssl: bool = False, requests_lib=requests):
        """Initializes a proxy object.

        :param base_uri: The base uri that is added before every api address. (For example: "https://localhost:8000/")
        :param service_account_token: The token that is added into the bearer authorization header.
        :param cacert: The file path to the file containing the ca cert that should verify the ssl connection.
        :param insecure_ssl: If this is true, ssl will be deactivated.
        :param requests_lib: The requests library wich makes the requests. Default: requests, but can be changed for mockups.
        """
        self.requests = requests_lib
        if service_account_token is None:
            logging.warning("No service account token.")
        if cacert is None:
            logging.warning("No cacert.")
        self.base_uri = base_uri
        self.service_account_token = service_account_token
        if insecure_ssl:
            self.verify = False
        elif cacert is None:
            self.verify = True
        else:
            self.verify = cacert

    def get(self, address: str) -> str:
        """Makes a get request.

        This method makes a get request to the Kubernetes API with authorization added and the SSL certificates
        checked.

        :param address: API path without base_uri. The address is put together with the base_uri.
        :return: The text body of the response. Should be in the YAML format.
        """
        headers = {"Authorization": f"Bearer {self.service_account_token}"}
        response = self.requests.get(self.base_uri + address,
                                     headers=headers, verify=self.verify)
        return response.text

    def post(self, address: str, data: str) -> str:
        """Makes a post request.

        This method makes a post request to the Kubernetes API with authorization added and the SSL certificates
        checked.

        :param address: API path without base_uri. The address is put together with the base_uri.
        :param data: POST body data. Should be a YAML string.
        :return: The text body of the response. Should be in the YAML format.
        """
        headers = {"Authorization": f"Bearer {self.service_account_token}",
                   "Content-Type": "application/yaml"}
        response = self.requests.post(self.base_uri + address,
                                      data=data, headers=headers, verify=self.verify)
        return response.text

    def delete(self, address) -> str:
        """Makes a delete request.

        This method makes a delete request to the Kubernetes API with authorization added and the SSL certificates
        checked.

        :param address: API path without base_uri. The address is put together with the base_uri.
        :return: The text body of the response. Should be in the YAML format.
        """
        headers = {"Authorization": f"Bearer {self.service_account_token}"}
        response = self.requests.delete(self.base_uri + address,
                                        headers=headers, verify=self.verify)
        return response.text


class APIRegistry:
    """This class is a container of Kubernetes API endpoints.

    Kubernetes API endpoints could be registered in the APIRegistry with the decorators `add_api_namespaced`,
    `add_api_not_namespaced` with a given name. After registering them they are available through their given name
    as attribute of this class. For example you register the Blahaj-API as "blahaj" then you can access it like this:
    `APIRegistry(proxy).blahaj`. This gives us a convenient way to add and access the needed Kubernetes APIs.
    """

    def __init__(self, proxy: Proxy):
        """Initializes an APIRegistry object.

        :param proxy: The proxy that should be used to make requests. The proxy already contains all information about
                      the Kubernetes API addresses.
        """
        self.proxy = proxy

    def __dir__(self):
        """This method is used to make the dynamic attributes of this class available to autocompletion.

        :return: A list of attributes this object has.
        """
        keys = set(super().__dir__())
        keys = keys.union(_API_EXTENSIONS_NAMESPACED.keys())
        keys = keys.union(_API_EXTENSIONS_NOT_NAMESPACED.keys())
        return list(keys)

    def __getattr__(self, name) -> Union['NamespacedApi', 'NotNamespacedApi']:
        """Executed on every attribute name.

        This method gets called on every attribute that is called on an APIRegistry object. This is used to dynamically
        add the API classes as attributes.

        :param name: Name of the attribute that was looked for.
        :return: An instance of the API class that belongs to the given attribute name.
        :raise: AttributeError: If attribute is not found.
        """
        cls: Union[Optional[Type['NamespacedApi']], Optional[Type['NotNamespacedApi']]]
        if cls := _API_EXTENSIONS_NAMESPACED.get(name):
            return cls(self.proxy)
        if cls := _API_EXTENSIONS_NOT_NAMESPACED.get(name):
            return cls(self.proxy)
        raise AttributeError(f'{name} not found')


class ApiExtension(ABC):
    """Used to extend the APIRegistry.

    To extend the APIRegistry, you need to implement either the NamespacedApi or NotNamespacedApi. Then add one
    of the decorators add_api_namespaced or add_api_not_namespaced or use these methods to add the classes manually.
    After that you can access an instance of the extension through the proxy as attributes for example
    "proxy.namespace" or "proxy.virtual_machine_instance". With that instances you are able to access the Kubernetes
    Api with the methods in NamespacedApi and NotNamespacedApi for example get_list or create.

    :param list_url: Will be formatted. Which variables will be inserted depends on the ApiExtension type.
    :param details_url: Will be formatted. Which variables will be inserted depends on the ApiExtension type.
    """

    list_url = None
    detail_url = None

    def __init__(self, proxy: Proxy):
        """Initializes an ApiExtension object.

        :param proxy: The proxy that should be used in this API to make requests.
        """
        self.proxy = proxy


class NamespacedApi(ApiExtension):
    """Abstract base class extension for resource object that are namespaced.

    :list_url: Will be formated with the variable "namespace".
    :detail_url: Will be formated with the variable "namespace" and "identifier".
    """

    def get_list(self, namespace: str) -> str:
        """Will get a list of all resource object in the namespace.

        :param namespace: The namespace where to get the list of resource object from.
        :return: A list of all resource object in the given namespace as YAML str.
        """
        return self.proxy.get(self.list_url.format(namespace=namespace))

    def create(self, namespace: str, data: str) -> str:
        """Creates a new resource object in the namespace.

        :param namespace: The namespace where the resource object should be created.
        :param data: The data of the resource object that is used to create it.
        :return: The newly added resource object as YAML string.
        """
        return self.proxy.post(self.list_url.format(namespace=namespace), data)

    def get(self, namespace: str, identifier: str) -> str:
        """Gets a specific resource object in the namespace.

        :param namespace: The namespace of the resource object.
        :param identifier: The identifier of the resource object.
        :return: The resource object as YAML str.
        """
        return self.proxy.get(self.detail_url.format(namespace=namespace, identifier=identifier))

    def delete(self, namespace: str, identifier: str) -> str:
        """Deletes a specific resource object in a namespace.

        :param namespace: The namespace of the resource object.
        :param identifier: The identifier of the resource object.
        :return: The deletion response.
        """
        return self.proxy.delete(self.detail_url.format(namespace=namespace, identifier=identifier))


class NotNamespacedApi(ApiExtension):
    """Abstract base class extension for resource object that are not namespaced.

    :list_url: Will not be formated and shouldn't contain variables.
    :detail_url: Will be formated with the variable "identifier".
    """

    def get_list(self) -> str:
        """Will get a list of all resource object.

        :return: A list of all resource object as YAML str.
        """
        return self.proxy.get(self.list_url)

    def create(self, data: str) -> str:
        """Creates a new resource object.

        :param data: The data of the resource object that is used to create it.
        :return: The newly added resource object as YAML string.
        """
        return self.proxy.post(self.list_url, data)

    def get(self, identifier: str) -> str:
        """Gets a specific resource object.

        :param identifier: The identifier of the resource object.
        :return: The resource object as YAML str.
        """
        return self.proxy.get(self.detail_url.format(identifier=identifier))

    def delete(self, identifier: str) -> str:
        """Deletes a specific resource object.

        :param identifier: The identifier of the resource object.
        :return: The deletion response.
        """
        return self.proxy.delete(self.detail_url.format(identifier=identifier))


@add_api_not_namespaced("namespace")
class Namespace(NotNamespacedApi):
    """Kubernetes Namespace API.

    The namespace api is used to create, get and delete namespaces in Kubernetes.
    """
    list_url = "/api/v1/namespaces"
    detail_url = "/api/v1/namespaces/{identifier}"


@add_api_namespaced("virtual_machine_instance")
class VirtualMachineInstance(NamespacedApi):
    """KubeVirt Virtual Machine Instance API.

    The virtual machine instance api is used to create, get and delete VMIs in Kubernetes. This only works when
    KubeVirt is installed in the Kubernetes cluster.
    """
    list_url = "/apis/kubevirt.io/v1alpha3/namespaces/{namespace}/virtualmachineinstances/"
    detail_url = "/apis/kubevirt.io/v1alpha3/namespaces/{namespace}/virtualmachineinstances/{identifier}"


@add_api_namespaced("network_policy")
class NetworkPolicy(NamespacedApi):
    """Kubernetes NetworkPolicy API.

    The network policy api is used to create, get and delete network policies in Kubernetes. This only works with a
    network plugin installed that implements network policies.
    """
    list_url = "/apis/networking.k8s.io/v1/namespaces/{namespace}/networkpolicies"
    detail_url = "/apis/networking.k8s.io/v1/namespaces/{namespace}/networkpolicies/{identifier}"



"""Contains a collection of controllers and a method to create all controllers at once."""

from dataclasses import dataclass

from lab_orchestrator_lib.controller.controller import NamespaceController, NetworkPolicyController, \
    DockerImageController, VirtualMachineInstanceController, LabController, LabInstanceController, UserController
from lab_orchestrator_lib.database.adapter import DockerImageAdapterInterface, LabInstanceAdapterInterface, \
    LabAdapterInterface, UserAdapterInterface
from lab_orchestrator_lib.kubernetes.api import APIRegistry


@dataclass
class ControllerCollection:
    """Contains all controllers."""
    user_ctrl: UserController
    namespace_ctrl: NamespaceController
    network_policy_ctrl: NetworkPolicyController
    docker_image_ctrl: DockerImageController
    virtual_machine_instance_ctrl: VirtualMachineInstanceController
    lab_ctrl: LabController
    lab_instance_ctrl: LabInstanceController


def create_controller_collection(
        registry: APIRegistry,
        user_adapter: UserAdapterInterface,
        docker_image_adapter: DockerImageAdapterInterface,
        lab_adapter: LabAdapterInterface,
        lab_instance_adapter: LabInstanceAdapterInterface,
        secret_key: str):
    """Initializes all controllers.

    :param registry: APIRegistry that should be injected into Kubernetes controllers.
    :param user_adapter: User adapter that should be injected into the controllers.
    :param docker_image_adapter: Docker image adapter that should be injected into the controllers.
    :param lab_adapter: Lab adapter that should be injected into the controllers.
    :param lab_instance_adapter: Lab instance adapter that should be injected into the controllers.
    :param secret_key: Secret key that should be used to create JWT tokens.
    :return: A controller collection with initialized controllers.
    """
    user_ctrl = UserController(user_adapter)
    namespace_ctrl = NamespaceController(registry)
    network_policy_ctrl = NetworkPolicyController(registry)
    docker_image_ctrl = DockerImageController(docker_image_adapter)
    virtual_machine_instance_ctrl = VirtualMachineInstanceController(registry, namespace_ctrl, docker_image_ctrl)
    lab_ctrl = LabController(lab_adapter)
    lab_instance_ctrl = LabInstanceController(
        adapter=lab_instance_adapter,
        virtual_machine_instance_ctrl=virtual_machine_instance_ctrl,
        namespace_ctrl=namespace_ctrl,
        lab_ctrl=lab_ctrl,
        network_policy_ctrl=network_policy_ctrl,
        user_ctrl=user_ctrl,
        secret_key=secret_key,
    )
    return ControllerCollection(
        user_ctrl=user_ctrl,
        namespace_ctrl=namespace_ctrl,
        network_policy_ctrl=network_policy_ctrl,
        docker_image_ctrl=docker_image_ctrl,
        virtual_machine_instance_ctrl=virtual_machine_instance_ctrl,
        lab_ctrl=lab_ctrl,
        lab_instance_ctrl=lab_instance_ctrl,
    )



"""Contains methods that are used to configure the Kubernetes API."""

import os
from dataclasses import dataclass
from typing import Optional

from lab_orchestrator_lib.kubernetes.api import Proxy, APIRegistry


@dataclass
class KubernetesConfig:
    """Configuration object that is used to create a Proxy and APIRegistry object.

    :arg service_account_token: Token that is used to authenticate against the Kubernetes API. This is probably the
                                token from a service account.
    :arg cacert: Ca certificate file that is used to secure the connection.
    :arg protocol: Either "https" or "http".
    :arg service_host: Host address of the Kubernetes API.
    :arg service_port: Port of the Kubernetes API.
    :arg base_uri: The base url that is used to connect to the Kubernetes API. (Combination of protocol, service_host and service_port)
    """
    service_account_token: Optional[str]
    cacert: Optional[str]
    protocol: str
    service_host: str
    service_port: str
    base_uri: str


def get_kubernetes_config():
    """Use this if you run the lib inside of Kubernetes.

    :return: A Kubernetes config that reads the token and ca cert from Kubernetes standard directories and it
    creates the other variables from Kubernetes variables.
    """
    with open('/var/run/secrets/kubernetes.io/serviceaccount/token', 'r') as file:
        service_account_token = file.read()
    cacert = '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt'
    protocol = "https"
    kubernetes_service_host = os.environ["KUBERNETES_SERVICE_HOST"]
    kubernetes_service_port = os.environ["KUBERNETES_SERVICE_PORT"]
    base_uri = f"{protocol}://{kubernetes_service_host}:{kubernetes_service_port}"
    kubernetes_config = KubernetesConfig(service_account_token, cacert, protocol, kubernetes_service_host,
                                         kubernetes_service_port, base_uri)
    return kubernetes_config


def get_development_config():
    """Use this if you run the lib inside of Kubernetes.

    Reads the environment variables `KUBERNETES_SERVICE_HOST` and `KUBERNETES_SERVICE_PORT`.

    :return: A Kubernetes config that disables SSL and assumes you are running `kubectl proxy`.
    """
    service_account_token = None
    cacert = None
    protocol = "http"
    kubernetes_service_host = os.environ["KUBERNETES_SERVICE_HOST"]
    kubernetes_service_port = os.environ["KUBERNETES_SERVICE_PORT"]
    base_uri = f"{protocol}://{kubernetes_service_host}:{kubernetes_service_port}"
    kubernetes_config = KubernetesConfig(service_account_token, cacert, protocol, kubernetes_service_host,
                                         kubernetes_service_port, base_uri)
    return kubernetes_config


def get_registry(kubernetes_config: KubernetesConfig):
    """Creates a Proxy and APIRegistry from the given Kuberntes_config.

    :param kubernetes_config: The Kubernetes config that should be used to create the proxy and api registry.
    :return: A APIRegistry that can be injected into Kubernetes controllers.
    """
    proxy = Proxy(kubernetes_config.base_uri, kubernetes_config.service_account_token, kubernetes_config.cacert)
    return APIRegistry(proxy)



"""Contains all adapters that needs to be implemented to use the lab orchestrator lib."""
from typing import List, Any, Dict

from lab_orchestrator_lib.model.model import DockerImage, Lab, LabInstance, Identifier, User


class UserAdapterInterface:
    """Adapter that is used to connect the user model to the database.

    Contains only get methods, because the lab_controller_lib won't create or delete users.
    """

    def get_all(self) -> List[User]:
        """Gives a list of all users.

        :return: A list of all users.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> User:
        """Gives a specific user.

        :param identifier: The identifier of the user.
        :return: A specific user.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()


class DockerImageAdapterInterface:
    """Adapter that is used to connect the docker image model to the database."""

    def create(self, name: str, description: str, url: str) -> DockerImage:
        """Creates a docker image and saves it to the database.

        :param name: Name of the docker image.
        :param description: Description of the docker image.
        :param url: Url of the docker image.
        :return: A newly added docker image.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_all(self) -> List[DockerImage]:
        """Gives all docker images.

        :return: A list of all docker images.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> DockerImage:
        """Gives a specific docker image.

        :param identifier: The identifier of the docker image.
        :return: The specific docker image.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_by_attr(self, attr: str, value: Any) -> DockerImage:
        """Gives a specific docker image.

        :param attr: The attribute name that should be used to filter.
        :param value: The value of the attribute that should be filtered.
        :return: The specific docker image.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def delete(self, identifier: Identifier) -> None:
        """Deletes a specific docker image.

        :param identifier: The identifier of the docker image.
        :return: None
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def save(self, obj: DockerImage) -> DockerImage:
        """Saves changes of the docker image to the database.

        :param obj: The docker image object that contains changes.
        :return: The docker image.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def filter(self, **kwargs: Dict[str, Any]) -> DockerImage:
        """Filters the docker images and returns the first docker image that matches the filter criteria.

        The database should be filtered by the attributes and belonging values that are given in the kwargs dictionary.

        :param kwargs: A dictionary with filters.
        :return: The first docker image that matches the filters.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()


class LabAdapterInterface:
    """Adapter that is used to connect the lab model to the database."""

    def create(self, name: str, namespace_prefix: str, description: str, docker_image_id: Identifier,
               docker_image_name: str) -> Lab:
        """Creates a lab and saves it to the database.

        :param name: Name of the lab.
        :param namespace_prefix: Namespace prefix of the lab.
        :param description: Description of the lab.
        :param docker_image_id: Docker image id of the lab.
        :param docker_image_name: Docker image name of the lab.
        :return: A newly added lab.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_all(self) -> List[Lab]:
        """Gives all labs.

        :return: A list of all labs.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> Lab:
        """Gives a specific lab.

        :param identifier: The identifier of the lab.
        :return: The specific lab.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_by_attr(self, attr: str, value: Any) -> Lab:
        """Gives a specific lab.

        :param attr: The attribute name that should be used to filter.
        :param value: The value of the attribute that should be filtered.
        :return: The specific lab.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def delete(self, identifier: Identifier) -> None:
        """Deletes a specific lab.

        :param identifier: The identifier of the lab.
        :return: None
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def save(self, obj: Lab) -> Lab:
        """Saves changes of the lab to the database.

        :param obj: The lab object that contains changes.
        :return: The lab.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def filter(self, **kwargs: Dict[str, Any]) -> Lab:
        """Filters the labs and returns the first lab that matches the filter criteria.

        The database should be filtered by the attributes and belonging values that are given in the kwargs dictionary.

        :param kwargs: A dictionary with filters.
        :return: The first lab that matches the filters.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()


class LabInstanceAdapterInterface:
    """Adapter that is used to connect the lab instance model to the database."""

    def create(self, lab_id: Identifier, user_id: Identifier) -> LabInstance:
        """Creates a lab instance and saves it to the database.

        :param lab_id: Lab id of the lab instance.
        :param user_id: User id of the lab instance.
        :return: A newly added lab instance.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_all(self) -> List[LabInstance]:
        """Gives all lab instances.

        :return: A list of all lab instances.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get(self, identifier: Identifier) -> LabInstance:
        """Gives a specific lab instance.

        :param identifier: The identifier of the lab instance.
        :return: The specific lab instance.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def get_by_attr(self, attr: str, value: Any) -> LabInstance:
        """Gives a specific lab instance.

        :param attr: The attribute name that should be used to filter.
        :param value: The value of the attribute that should be filtered.
        :return: The specific lab instance.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def delete(self, identifier: Identifier) -> None:
        """Deletes a specific lab instance.

        :param identifier: The identifier of the lab instance.
        :return: None
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def save(self, obj: LabInstance) -> LabInstance:
        """Saves changes of the lab instance to the database.

        :param obj: The lab instance object that contains changes.
        :return: The lab instance.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()

    def filter(self, **kwargs: Dict[str, Any]) -> LabInstance:
        """Filters the lab instances and returns the first lab instance that matches the filter criteria.

        The database should be filtered by the attributes and belonging values that are given in the kwargs dictionary.

        :param kwargs: A dictionary with filters.
        :return: The first lab instance that matches the filters.
        :raise NotImplementedError: Method needs to be implemented.
        """
        raise NotImplementedError()
